(()=>{(()=>{let e={},t=[{name:"_refresh",endpoint:["/tetra/tetra/default/router/_refresh"]},{name:"navigate",endpoint:["/tetra/tetra/default/router/navigate"]}],r=["key","current_component","current_path"],n="tetra__default__router";window.document.addEventListener("alpine:init",()=>{Tetra.makeAlpineComponent(n,e,t,r)})})();(()=>{let e={},t=[{name:"_refresh",endpoint:["/tetra/tetra/default/link/_refresh"]},{name:"click",endpoint:["/tetra/tetra/default/link/click"]}],r=["key"],n="tetra__default__link";window.document.addEventListener("alpine:init",()=>{Tetra.makeAlpineComponent(n,e,t,r)})})();(()=>{let e={},t=[{name:"_refresh",endpoint:["/tetra/tetra/default/redirect/_refresh"]}],r=["key"],n="tetra__default__redirect";window.document.addEventListener("alpine:init",()=>{Tetra.makeAlpineComponent(n,e,t,r)})})();})();
//# sourceMappingURL=tetra_default-LHYUSLYH.js.map
