{% load static %}
{% if csrf_token %}
<script>
  window.__tetra_csrfToken = '{{ csrf_token }}';
  window.__tetra_staticRoot = '{% get_static_prefix %}';
  window.__tetra_useWebsockets = {{ use_websockets|yesno:"true,false" }};
</script>
{% endif %}
{% for lib in libs %}
<script defer src="{{ lib.js_url }}"></script>
{% endfor %}
<script defer src="{% static 'tetra/js/tetra.js' %}"></script>
{% if include_alpine %}
  {% if debug|yesno:"true,false" %}
    <script defer src="{% static 'tetra/js/alpinejs.morph.cdn.js' %}"></script>
    <script defer src="{% static 'tetra/js/alpinejs.cdn.js' %}"></script>
  {% else %}
    <script defer src="{% static 'tetra/js/alpinejs.morph.cdn.min.js' %}"></script>
    <script defer src="{% static 'tetra/js/alpinejs.cdn.min.js' %}"></script>
  {% endif %}
{% endif %}