(()=>{(()=>{let e={},t=[{name:"_refresh",endpoint:["/__tetra__tetra/test_router/home/_refresh"]}],r=["key"],_="tetra__test_router__home";window.document.addEventListener("alpine:init",()=>{Tetra.makeAlpineComponent(_,e,t,r)})})();(()=>{let e={},t=[{name:"_refresh",endpoint:["/__tetra__tetra/test_router/about/_refresh"]}],r=["key"],_="tetra__test_router__about";window.document.addEventListener("alpine:init",()=>{Tetra.makeAlpineComponent(_,e,t,r)})})();(()=>{let e={},t=[{name:"_refresh",endpoint:["/__tetra__tetra/test_router/my_router/_refresh"]},{name:"navigate",endpoint:["/__tetra__tetra/test_router/my_router/navigate"]}],r=["key","current_component","current_path"],_="tetra__test_router__my_router";window.document.addEventListener("alpine:init",()=>{Tetra.makeAlpineComponent(_,e,t,r)})})();})();
//# sourceMappingURL=tetra_test_router-JNWYUKEM.js.map
